///////////////////////////////////////////////////////////////////////////////
// Serialize - Output a machine-readable serialization of Java objects
// Copyright (C) 2002, 2004 C. Brian Jones <cbj@gnu.org>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
///////////////////////////////////////////////////////////////////////////////

package com.oryxsoft.test.serialization;

import gnu.getopt.Getopt;
//import JSX.ObjectWriter;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.LineNumberReader;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.ObjectOutputStream;
import java.io.NotSerializableException;

/**
 * This program can be used to output serialized objects to 
 * individual files for each class.  The files are created
 * through the normal serialization process for Java and are
 * either the typical binary format or XML.
 * <p>
 * As XML serialization has not been standardized 
 * or made available in early JDKs, you can use JSX to output
 * an XML file containing the serialized object's data as the 
 * Java serialization mechanism would have written it in binary
 * form.  This can be used to diagnose why two binary serializations
 * are different and can be used in addition to or instead of the
 * special serialized information only recently added to the JDK
 * documentation.
 *
 * @author C. Brian Jones &lt;<a href="mailto:cbj@gnu.org">cbj@gnu.org</a>&gt;
 */
public class Serialize
{
  private static String _outputDirectory = ".";
  private static boolean _xml = false;
  
  public static void main (String[] argv)
  {
    if (argv.length < 1) 
      help ();

    Getopt g = new Getopt ("Serialize", argv, "d:x");

    int c;
    String arg;
    while ((c = g.getopt ()) != -1)
    {
      switch (c)
      {
        case 'd':
          _outputDirectory = g.getOptarg ();
          break;
        case 'x':
          _xml = true;
          break;
        case '?':
          help ();
          break;
        default:
          help ();
          break;
      }
    }
    
    Serialize ser = new Serialize ();
    
    for (int i = g.getOptind(); i < argv.length ; i++)
    {
      File f = new File (argv[i]);
      if (f.exists ())
        ser.readFile (f);
      else
        ser.writeClass (argv[i]);
    }

    // In case AWT threads do not die appropriately
    System.exit (0);
  }
  
  private void readFile (File f)
  {
    try
    {
      LineNumberReader reader = new LineNumberReader (new BufferedReader (new FileReader (f)));
      String cls;
      while ((cls = reader.readLine ()) != null)
        writeClass (cls);
      reader.close ();
    }
    catch (Throwable t) 
    {
      System.err.println (t.getMessage ());
      System.exit (1);
    }
  }

  private Class _serial = null;
  private ObjectOutputStream _objout = null;  
  private File _file = null;
  
  private void writeClass (String clsName)
  {
    System.out.print (clsName);
    
    try
    {
      String filePrefix = clsName.replace ('.', File.separatorChar);
      StringBuffer buf = new StringBuffer (_outputDirectory);
      buf.append (File.separator);
      if (filePrefix.indexOf (File.separator) != -1)
        buf.append (filePrefix.substring (0, filePrefix.lastIndexOf (File.separator)));
      else
        buf.append (filePrefix);
      String dirPrefix = buf.toString ();
      File dir = new File (dirPrefix);
      if (! dir.isDirectory ())
        dir.mkdirs ();
    
      if (_xml)
      {
        throw new RuntimeException("JSX code commented out, non-free software cannot be distributed now.");
//        _file = new File (_outputDirectory, filePrefix + Constants.XML_TYPE);
//        _objout = new ObjectWriter (new BufferedOutputStream (new FileOutputStream (_file)));
      }
      else
      {
        _file = new File (_outputDirectory, filePrefix + Constants.BINARY_TYPE);        
        _objout = new ObjectOutputStream (new BufferedOutputStream (new FileOutputStream (_file)));
      }
      if (_serial == null)
      {
        _serial = Class.forName ("java.io.Serializable");
      }
    }
    catch (Throwable t)
    {
      System.err.println (": error, " + t.getMessage ());
      t.printStackTrace ();
      close (true);
//      System.exit (1);
    }
      
    Object obj = null;
    boolean writeObj = false;
    try
    {
      Class cls = Class.forName (clsName);
      if (cls.isInterface ())
      {
        System.out.println (": ok, is interface");
        return;
      }
      int flags = cls.getModifiers ();
      if (! Modifier.isPublic (flags))
      {
        System.out.println (": ok, not public");
        return;
      }
      if (Modifier.isAbstract (flags))
      {
        System.out.println (": ok, abstract");
        return;
      }
      obj = getInstance (cls);
      if (_serial.isInstance (obj))
        writeObj = true;
    }
    catch (ClassNotFoundException cnfe)
    {
      System.out.println (": fail, class not found");
      close (true);
      return;
    }
    catch (UnsupportedOperationException e)
    {
      System.out.println (": fail, no instance");
      close (true);
      return;
    }
    catch (NoClassDefFoundError e)
    {
      System.out.println (": fail, class not found");
      close (true);
      return;
    }
    catch (Throwable t) 
    {
      System.out.println (": fail, " + t.getClass ().getName () + " : " + t.getMessage ());
//      t.printStackTrace (System.out);
      close (true);
      return;
    }
    
    try
    {
      if (writeObj)
      {
        _objout.writeObject (obj);
      }
    }
    catch (NotSerializableException e)
    {
      System.out.println (": ok, not serializable");
      close (true);
      return;
    }
    catch (SecurityException e)
    {
      System.out.println (": ok, security exception");
      close (true);
      return;
    }
//    catch (org.omg.CORBA.BAD_OPERATION oe)
//    {
//      System.out.println (": ok, bad operation");
//      close (true);
//      return;
//    }
    catch (NoClassDefFoundError e)
    {
      System.out.println (": fail, class not found");
      close (true);
      return;
    }
    catch (Throwable t)
    {
      System.out.println (": error, " + t.getMessage ());
//      t.printStackTrace ();
      close (true);
      return;
    }
    if (writeObj)
      System.out.println (": ok");
    else
      System.out.println (": ok, not serializable");

    close ();
  }

  private Object getInstance (Class cls)
  {
    try
    {
      Object obj = cls.newInstance ();
      return obj;
    }
    catch (Throwable t) { }

    try
    {
      Object obj = getInstanceSpecialClasses (cls);
      if (obj != null) 
        return obj;
    }
    catch (Throwable t) { }

    // Try to find a simple constructor using primitive arguments
    Constructor[] constructors = cls.getConstructors ();    
    for (int index = 0; index < constructors.length; index++)
    {
      int flags = constructors[index].getModifiers ();
      if (Modifier.isPublic (flags))
      {
        Class[] params = constructors[index].getParameterTypes ();
        // Avoid constructors with parameters with the same type as cls
        boolean ignoreConstructor = false;
        for (int j = 0; j < params.length; j++)
        {
          if (cls.isAssignableFrom (params[j]))
          {
            ignoreConstructor = true;
            break;
          }
        }
        if (ignoreConstructor)
          continue;

        try
        {
          Object[] initargs = getParameters (params);
          Object obj = constructors[index].newInstance (initargs);
          return obj;
        }
        catch (Throwable t)
        {
          continue;
        }
      }
    }
    
    throw new UnsupportedOperationException ("unable to create instance");
  }

  private Object[] getParameters (Class[] params)
  {
    Object[] initargs = new Object[params.length];
    boolean paramsOkay = true;
    for (int j = 0; j < params.length; j++)
    {
      if (isPrimitive (params[j]))
        initargs[j] = getPrimitiveValue (params[j]);
      else
        initargs[j] = getInstance (params[j]);
    }
    return initargs;
  }
  
  private boolean isPrimitive (Class c)
  {
    if (c.isPrimitive ())
      return true;
    if (c.isAssignableFrom (java.lang.String.class))
      return true;
    return false;
  }

  private Object getPrimitiveValue (Class c)
  {
    if (c.isAssignableFrom (java.lang.Boolean.TYPE))
      return Boolean.TRUE;
    else if (c.isAssignableFrom (java.lang.Byte.TYPE))
      return new Byte ((byte)0);
    else if (c.isAssignableFrom (java.lang.Character.TYPE))
      return new Character ('a');
    else if (c.isAssignableFrom (java.lang.Short.TYPE))
      return new Short ((short)0);
    else if (c.isAssignableFrom (java.lang.Integer.TYPE))
      return new Integer (0);
    else if (c.isAssignableFrom (java.lang.Long.TYPE))
      return new Long (0);
    else if (c.isAssignableFrom (java.lang.Float.TYPE))
      return new Float (0.0f);
    else if (c.isAssignableFrom (java.lang.Double.TYPE))
      return new Double (0.0d);
    else if (c.isAssignableFrom (java.lang.String.class))
      return new String ("a");
    return null;
  }

  private Object getInstanceSpecialClasses (Class cls)
  {
    String clsName = cls.getName();
    try
    {
      if (clsName.equals("java.awt.ComponentOrientation"))
      {
        Field field = cls.getField("LEFT_TO_RIGHT");
        return field.get(null);
      }
      else if (clsName.equals("java.awt.Font"))
      {
        Method method = cls.getDeclaredMethod("decode", new Class[]{Class.forName("java.lang.String")});
        return method.invoke(null, new Object[]{null});
      }
      // FIXME FAIL?
      else if (clsName.equals("java.awt.MediaTracker"))
      {
        Object applet = getInstance(Class.forName("java.applet.Applet"));
        Constructor constructor = cls.getConstructor(new Class[]{Class.forName("java.applet.Applet")});
        return constructor.newInstance(new Object[]{applet});
      }
      else if (clsName.equals("java.awt.SystemColor"))
      {
        Field field = cls.getField("desktop");
        return field.get(null);
      }
      else if (clsName.equals("java.awt.TextArea"))
      {
      }
      else if (clsName.equals("java.awt.TextComponent"))
      {
      }
      else if (clsName.equals("java.awt.TextField"))
      {
      }
      else if (clsName.equals("java.awt.Window"))
      {
        Object frame = getInstance(Class.forName("java.awt.Frame"));
        Constructor constructor = cls.getConstructor(new Class[]{Class.forName("java.awt.Frame")});
        return constructor.newInstance(new Object[]{frame});
      }
      else if (clsName.equals("java.awt.color.ICC_ColorSpace"))
      {
      }
      else if (clsName.equals("java.awt.color.ICC_Profile"))
      {
      }
      else if (clsName.equals("java.awt.color.ICC_ProfileGray"))
      {
      }
      else if (clsName.equals("java.awt.color.ICC_ProfileRGB"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DragGestureEvent"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DragSource"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DragSourceContext"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DragSourceDragEvent"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DragSourceDropEvent"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DragSourceEvent"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DropTarget"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DropTargetContext"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DropTargetDragEvent"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DropTargetDropEvent"))
      {
      }
      else if (clsName.equals("java.awt.dnd.DropTargetEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.AdjustmentEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.ComponentEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.ContainerEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.FocusEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.InputMethodEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.InvocationEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.ItemEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.KeyEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.MouseEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.PaintEvent"))
      {
      }
      else if (clsName.equals("java.awt.event.WindowEvent"))
      {
      }
      else if (clsName.equals("java.awt.font.TextAttribute"))
      {
      }
      else if (clsName.equals("java.beans.beancontext.BeanContextMembershipEvent"))
      {
      }
      else if (clsName.equals("java.beans.beancontext.BeanContextServiceAvailableEvent"))
      {
      }
      else if (clsName.equals("java.beans.beancontext.BeanContextServiceRevokedEvent"))
      {
      }
      else if (clsName.equals("java.io.ObjectStreamClass"))
      {
        Method method = cls.getDeclaredMethod("lookup", new Class[]{Class.forName("java.lang.Class")});
        return method.invoke(null, new Object[]{Class.forName("java.lang.String")});
      }
      else if (clsName.equals("java.io.OptionalDataException"))
      {
      }
      // If you don't have Class and String then all hell has already broken
      else if (clsName.equals("java.lang.Class"))
      {
        return Class.forName("java.lang.String");
      }
      else if (clsName.equals("java.net.InetAddress"))
      {
        Method method = cls.getDeclaredMethod("getLocalHost", new Class[]{});
        return method.invoke(null, new Object[]{});
      }
      else if (clsName.equals("java.net.URL"))
      {
        Constructor constructor = cls.getConstructor(new Class[]{Class.forName("java.lang.String")});
        return constructor.newInstance(new Object[]{"file:///"});
      }
      else if (clsName.equals("java.rmi.activation.ActivationDesc"))
      {
      }
      else if (clsName.equals("java.rmi.activation.ActivationGroup_Stub"))
      {
      }
      else if (clsName.equals("java.rmi.activation.ActivationGroupDesc"))
      {
      }
      else if (clsName.equals("java.rmi.activation.ActivationGroupDesc$CommandEnvironment"))
      {
      }
      else if (clsName.equals("java.rmi.activation.ActivationGroupID"))
      {
      }
      else if (clsName.equals("java.rmi.activation.ActivationID"))
      {
      }
      else if (clsName.equals("java.rmi.server.UnicastRemoteObject"))
      {
      }
      else if (clsName.equals("java.security.CodeSource"))
      {
      }
      else if (clsName.equals("java.security.GuardedObject"))
      {
      }
      else if (clsName.equals("java.security.KeyPair"))
      {
      }
      else if (clsName.equals("java.security.SignedObject"))
      {
      }
      else if (clsName.equals("java.security.UnresolvedPermission"))
      {
      }
      else if (clsName.equals("java.text.AttributedCharacterIterator$Attribute"))
      {
      }
    }
    catch (Throwable t) { }
    return null; 
  }
  
  private void close (boolean delete)
  {    
    if (_objout != null)
    {
      try
      {
        _objout.flush ();
        _objout.close ();
      }
      catch (Throwable t) { }    
      try
      {
        if (delete)
          if (_file != null)
            _file.delete ();
        
      }
      catch (Throwable t) { }
    }
  }

  private void close ()
  {
    close (false);
  }
  
  private static void help ()
  {
    System.err.println ("Serialize " + Constants.VERSION + ", (c) 2002, 2004 C. Brian Jones");
    System.err.println ("Usage: serialize [-dx] [FILE|CLASS]");
    System.err.println ("  -d <directory>    output directory");
    System.err.println ("  -x                xml output, default is binary");
    System.exit (1);
  }
}

